import errno
import importlib
import logging
from typing import List

from addict import addict

from kodexa import Assistant, AssistantResponse
from kodexa import ContentEvent, ContentNode, Document, DocumentActor, DocumentTransition, TransitionType
from kodexa.assistant.assistant import AssistantMetadata
from kodexa.model import AssistantEvent, ActorType
from kodexa.model.objects import ScheduledEvent, ExceptionDetails
from kodexa.platform.client import DocumentStoreEndpoint

logger = logging.getLogger()


class DocumentTestCaptureStep:

    def __init__(self):
        self.documents = []

    def process(self, document):
        self.documents.append(document)
        return document


def simplify_node(node: ContentNode):
    """

    Args:
      node: ContentNode:

    Returns:

    """
    return {
        "index": node.index,
        "node_type": node.node_type,
        "features": [feature.to_dict() for feature in node.get_features()],
        "content": node.content,
        "content_parts": node.get_content_parts(),
        "children": [simplify_node(child_node) for child_node in node.get_children()]
    }


def simplify_document(document: Document) -> dict:
    """
    This method can be used to simplify the structure when we want to capture it for
    comparison in testing

    Args:
      document: Document: the document to simplify

    Returns:
        A dictionary based simplified representation

    """
    return {
        "content_node": simplify_node(document.get_root())
    }


def compare_document(document: Document, filename: str, throw_exception=True):
    """

    Args:
      document: Document:
      filename: str:
      throw_exception:  (Default value = True)

    Returns:

    """
    from os import path
    import json
    import os

    try:
        os.makedirs('test_snapshots')
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    filename = "test_snapshots/" + filename

    if not path.exists(filename):
        with open(filename, 'w') as f:
            simplified_document = simplify_document(document)
            json.dump(simplified_document, f)

        logger.warning("WARNING!!! Creating snapshot file")
        raise Exception("Creating snapshot, invalid test")

    with open(filename) as f:
        snapshot_document = json.load(f)

    target_document = json.loads(json.dumps(simplify_document(document)))

    from deepdiff import DeepDiff
    diff = DeepDiff(snapshot_document, target_document, ignore_order=False)

    if bool(diff) and throw_exception:
        print(diff)
        raise Exception('Document does not match')

    return diff


class AssistantTestHarness:
    """
    A test harness to allow the testing of assistants in unit tests and offline for development

    >>> util = ExtensionPackUtil("../kodexa.yml")
    >>> harness = util.get_assistant("my-assistant", stores=[my_local_store])

    """

    def __init__(self, assistant: Assistant, stores: List[DocumentStoreEndpoint], kodexa_metadata_path: str,
                 metadata: AssistantMetadata, content_provider=None):
        """
        Initialize the test harness

        Args:
            assistant: the instance of the assistant
            stores: the list of stores (usually LocalDocumentStore) that we will use to monitor for events
            kodexa_metadata_path (str): the path to the kodexa.yml (or kodexa.json)
            metadata (AssistantMetadata): the assistant metadata to use for this assistant
            content_provider: optional provider for content
        """
        self.assistant = assistant
        self.stores: List[DocumentStoreEndpoint] = stores
        self.kodexa_metadata_path = kodexa_metadata_path
        self.metadata = metadata
        self.content_provider = content_provider

        for store in self.stores:
            store.register_listener(self)

    def test_assistant_event(self, assistant_event: AssistantEvent) -> AssistantResponse:
        """
        Pass an assistant event into the assistant

        :param assistant_event: the assistant event to process
        :return: the response from the event
        """
        from kodexa import AssistantContext
        assistant_context = AssistantContext(self.metadata, self.kodexa_metadata_path, self.stores,
                                             self.content_provider)

        return self.assistant.process_event(assistant_event, assistant_context)

    def test_scheduled_event(self, scheduled_event: ScheduledEvent) -> AssistantResponse:
        """
        Pass a scheduled event into the assistant

        :param scheduled_event: the scheduled event to process
        :return: the response from the assistant
        """
        from kodexa import AssistantContext
        assistant_context = AssistantContext(self.metadata, self.kodexa_metadata_path, self.stores,
                                             self.content_provider)

        return self.assistant.process_event(scheduled_event, assistant_context)

    def process_event(self, event: ContentEvent):
        """
        The harness will take the content event and
        will pass it to the assistant - then we will
        take each of the pipelines and run the document
        through them in turn (note in the platform this might be in parallel)

        Args:
          event: ContentEvent: The event to process

        Returns:
          None

        """
        from kodexa import AssistantContext
        assistant_context = AssistantContext(self.metadata, self.kodexa_metadata_path, self.stores)

        response: AssistantResponse = self.assistant.process_event(event, assistant_context)

        # We need to get the document down
        store = self.get_store(event)

        for assistant_pipeline in response.pipelines:
            document = store.get_document_by_content_object(event.document_family, event.content_object)
            if document is not None:
                pipeline = assistant_pipeline.pipeline
                from kodexa.model.model import ContentObjectReference
                pipeline.connector = [
                    ContentObjectReference(content_object=event.content_object, document=document, store=store,
                                           document_family=event.document_family)]
                pipeline_context = pipeline.run()

                if pipeline_context.output_document is not None and assistant_pipeline.write_back_to_store:
                    # We need to build the transition between the old and the new
                    document_relationship = DocumentTransition()
                    document_relationship.transition_type = TransitionType.derived
                    document_relationship.source_content_object_id = event.content_object.id
                    document_relationship.actor = DocumentActor(actorId="testing", actorType=ActorType.assistant)

                    store.add_related_document_to_family(event.document_family.id, document_relationship,
                                                         pipeline_context.output_document)

    def get_store(self, event: ContentEvent) -> DocumentStoreEndpoint:
        """
        Get a document store for the event (based on the document family ID)

        Args:
          event: ContentEvent:

        Returns:
          The instance of the document store
        """
        for store in self.stores:
            if event.document_family.store_ref == store.ref:
                return store

        raise Exception(f"Unable to get store ref {event.document_family.store_ref}")


class OptionException(Exception):
    """
    An exception that is raised when there is a problem with a requests option
    """


class ExceptionBuilder:
    """
    A helper to build an exception details from the last exception
    """

    @staticmethod
    def build_exception_details():
        import sys
        import better_exceptions
        et, ev, tb = sys.exc_info()
        return ExceptionDetails(**{'errorType': et.__name__, 'errorMessage': str(ev),
                                                'message': "An unexpected exception has occurred",
                                                'help': "\n".join(
                                                    better_exceptions.format_exception(*sys.exc_info()))})


class ExtensionPackUtil:
    """
    A utility that can be used to access an action defined in a kodexa.yml.

    This allows you to use the kodexa.yml in unit tests to ensure it matches your current action code

    >>> from kodexa import *
    >>> util = ExtensionPackUtil("../kodexa.yml")
    >>> pipeline = Pipeline()
    >>> pipeline.add_step(util.get_step("my-action",{"my-option": "cheese"}))
    >>> pipeline.run()

    """

    def __init__(self, file_path='kodexa.yml'):
        self.file_path = file_path

        if file_path.endswith('.yml'):
            import yaml

            with open(file_path, 'r') as stream:
                self.kodexa_metadata = addict.Dict(yaml.safe_load(stream))

        if file_path.endswith('.json'):
            import json

            with open(file_path, 'r') as stream:
                self.kodexa_metadata = addict.Dict(json.load(stream))

    def get_step(self, action_slug, options=None):
        """

        Args:
          action_slug (str): the slug to the action (ie. pdf-parser)
          options (dict):  the options for the action as a dictionary (Default value = None)

        Returns:

        """
        if options is None:
            options = {}

        for service in self.kodexa_metadata.services:
            if service.type == 'action' and service.slug == action_slug:
                # TODO We need to validate all the options

                if len(service.metadata.options) > 0:
                    option_names = []
                    for option in service.metadata.options:
                        option_names.append(option.name)
                        if option.name not in options and option.default is not None:
                            options[option.name] = option.default
                        if option.required and option.name not in options:
                            raise OptionException(f"Missing required option {option.name}")

                    for option_name in options.keys():
                        if option_name not in option_names:

                            # We need to determine if this is actually a group
                            is_group = False
                            for check_option in service.metadata.options:
                                if check_option['group'] is not None:
                                    if check_option['group']['name'] == option_name:
                                        is_group = True

                            if not is_group:
                                raise OptionException(f"Unexpected option {option_name}")

                # We need to create and return our action
                module = importlib.import_module(service.step.package)
                klass = getattr(module, service.step['class'])
                new_instance = klass(**options)

                # Since we will be using to access metadata we will need to
                # make sure we have a to_dict() that is able to convert this step properly

                import types

                def general_to_dict(self):
                    return {
                        'ref': f'./{action_slug}',
                        'options': options
                    }

                new_instance.to_dict = types.MethodType(general_to_dict, new_instance)
                return new_instance

        raise Exception("Unable to find the action " + action_slug)

    def get_assistant_test_harness(self, assistant_slug, assistant_metadata: AssistantMetadata, options=None,
                                   stores=None) -> AssistantTestHarness:
        """Provides a local test harness that can be used to validate the functionality
        of an assistant in a test case

        Args:
          assistant_slug (str): The slug for the assistant (ie. pdf-parser-assistant)
          assistant_metadata (AssistantMetadata): the metadata to use for the assistant
          stores (List): a list of the document stores to monitor (Default value = None)
          options (dict): The options to provide  (Default value = None)

        Returns:
          The assistant test harness

        """
        if stores is None:
            stores = []
        assistant = self.get_assistant(assistant_slug, options)

        return AssistantTestHarness(assistant, stores, self.file_path, assistant_metadata)

    def get_assistant(self, assistant_slug, options=None):
        """Create an instance of an assistant from the Kodexa metadata

        Args:
          assistant_slug: param options:
          options:  (Default value = None)

        Returns:

        """
        if options is None:
            options = {}

        for service in self.kodexa_metadata.services:
            if service.type == 'assistant' and service.slug == assistant_slug:
                # TODO We need to validate all the options

                # We need to create and return our action

                logger.info(f"Creating new assistant {service.assistant}")
                module = importlib.import_module(service.assistant.package)
                klass = getattr(module, service.assistant['class'])
                return klass(**options)

        raise Exception("Unable to find the assistant " + assistant_slug)



import re

from kodexa import ContentNode
from kodexa.spatial.bbox_common import overlaps_with, width_of_overlap, percent_nodes_overlap
from typing import Optional, List, Dict, Tuple
import logging

logger = logging.getLogger()

OVERLAP_PERCENTAGE = 0.4
MIN_OVERLAP_PERCENTAGE_X = 0.2
MIN_OVERLAP_PERCENTAGE_Y = 0.4
COL_SPACE_MULTIPLIER = 2.0


def transform_line_to_columns(node, col_space_multiplier=3.0, col_marker_line=None,
                              use_graphical_nodes=False, graphic_slop=1.0):
    """
    Calculate the potential columns under this node and then transform the line to have
    a new set of nodes (type column).
    Note that this action will transform the document.

    :param node: The node that we want to transform to columns
    :param col_space_multiplier: Number of spaces between columns relative to the mean width of
    the characters on the page. Default is 3.0.
    :param col_marker_line: Line that dictates the positions of the columns. Default is None - which means the columns
    will be identified based on the spatial values of the nodes.
    :param use_graphical_nodes: If set to True, the graphical figures (lines and rects)
    that are identified on the page will be used to identify the columns of the table. Default is False.
    :param graphic_slop: The value of the slop allowed in lining up the nodes.
    Default is 1.0 which means it is +/- 1.0 of the mean character width of the line.
    This is only checked when use_graphical_nodes is set to True.

    """

    if node.node_type != 'line':
        return

    if node.get_children()[0].node_type == 'column':
        # This line has already been transformed to columns
        return

    # Grab graphic elements on the page
    page = node.select_first('parent::page')

    line_words = node.get_children().copy()
    mean_width = node.get_statistics()['updated_mean_width']

    # If col_markers is not None, use it to identify the words that belong to a column.
    # We identify the spaces by checking the overlap in the spaces
    insert_empty_col = False
    if col_marker_line:
        col_markers = col_marker_line.get_children().copy()
        column_words_index = []
        for cm_idx, cm in enumerate(list(col_markers)[0:-1]):
            # We are checking where the spaces overlap
            cm_bbox = cm.get_bbox()
            next_cm_bbox = col_markers[cm_idx + 1].get_bbox()

            # Identify the spaces that overlap and that the x2 of the word is after the next column_markers x
            gap_index_list = [i for i in range(1, len(line_words))
                              if (line_words[i - 1].get_bbox()[2] <= next_cm_bbox[0] < line_words[i].get_bbox()[2] and
                                  line_words[i].get_x() >= cm_bbox[2])]

            if len(gap_index_list) > 0:
                column_words_index.append(max(gap_index_list))
            else:
                # Identify the words that overlap - this is true for those columns that are 'inserted'
                # The 'overlap' should be 100% (meaning the words - with text - should be fully overlapping
                # with the empty-text column
                word_index_list = [i + 1 for i in range(0, len(line_words)) if
                                   (cm.get_all_content() == '' and cm_bbox[0] <= line_words[i].get_x() <
                                    line_words[i].get_bbox()[2] <= cm_bbox[2]) or
                                   (len(cm.get_all_content()) > 0 and
                                    (overlaps_with(line_words[i], cm) or
                                     line_words[i].get_bbox()[2] < next_cm_bbox[0]))]

                if len(word_index_list) > 0:
                    column_words_index.append(max(word_index_list))

    else:
        column_words_index = []

        for i in range(1, len(line_words)):
            is_break = False

            if use_graphical_nodes:
                is_break = check_graphical_nodes_break(page, graphic_slop, mean_width, line_words, i)

            elif line_words[i].get_x() - (line_words[i - 1].get_x() + line_words[i - 1].get_width()) >= \
                    col_space_multiplier * mean_width:
                is_break = True

            if is_break:
                column_words_index.append(i)

    # Insert 0 as the first index for column words; and len(node.words) as last
    column_words_index.insert(len(column_words_index), len(line_words))
    if 0 not in column_words_index:
        column_words_index.insert(0, 0)

    num_columns = 0
    new_columns = []
    for c_idx, n_idx in zip(column_words_index[:-1], column_words_index[1:]):
        column_words = []

        for i in range(c_idx, n_idx):
            column_words.append(line_words[i])

        # Create a new Column if there are column words
        if len(column_words) > 0:
            column_node = create_column_node_from_words(node.document, node, column_words)
            column_node.set_bbox_from_children()
            num_columns += 1
            new_columns.append(column_node)

    node.adopt_children(new_columns, replace=True)


def check_graphical_nodes_break(node, graphic_slop, mean_width, line_words, i):
    is_break = False
    graphical_nodes = node.select('//rect | //figure-line')

    x1 = (line_words[i - 1].get_x() + line_words[i - 1].get_width()) - (graphic_slop * mean_width)
    x2 = (line_words[i].get_x()) + (graphic_slop * mean_width)
    for node in graphical_nodes:

        # if line_words[i].get_y() + line_words[i].get_height() - (graphic_slop * mean_width) \
        #         <= node.get_y() + node.get_height() and \
        #         line_words[i].get_y() >= node.get_y() + (graphic_slop * mean_width):

        if x1 < node.get_bbox()[0] < x2:
            is_break = True
            original_bbox = line_words[i].get_bbox()
            original_bbox = [node.get_bbox()[0], original_bbox[1], original_bbox[2], original_bbox[3]]
            line_words[i].set_bbox(original_bbox)

            previous_bbox = line_words[i - 1].get_bbox()
            previous_bbox = [previous_bbox[0], previous_bbox[1], node.get_bbox()[0] - 0.01,
                             previous_bbox[3]]
            line_words[i - 1].set_bbox(previous_bbox)
        if x1 < node.get_bbox()[2] < x2:
            is_break = True
            original_bbox = line_words[i].get_bbox()
            original_bbox = [node.get_bbox()[2], original_bbox[1], original_bbox[2], original_bbox[3]]
            line_words[i].set_bbox(original_bbox)

            previous_bbox = line_words[i - 1].get_bbox()
            previous_bbox = [previous_bbox[0], previous_bbox[1], node.get_bbox()[2] - 0.01,
                             previous_bbox[3]]
            line_words[i - 1].set_bbox(previous_bbox)

    return is_break


def create_column_node_from_words(document, line, column_words):
    column_node = document.create_node(node_type='column')

    cw_first_bbox = column_words[0].get_bbox()
    cw_last_bbox = column_words[-1].get_bbox()
    column_node.set_bbox([cw_first_bbox[0], cw_first_bbox[1],
                          cw_last_bbox[2], cw_first_bbox[3]])

    column_node.adopt_children(column_words, replace=True)

    return column_node


def to_table(node, tag_name,
             col_space_multiplier=3.0, col_marker_line=None,
             insert_col_before=False, insert_col_after=False, insert_col_index=None,
             use_graphical_nodes=False, graphic_slop=1.0):
    """
    Uses a tag name to convert matching child nodes into tables
    If col_marker_line is given, it will be used to identify the number of columns
    in the table, with the col_space_multiplier.

    Note that this transforms the document - adding 'columns' as children of 'lines'.

    :param tag_name: The tag name used for the table rows
    :param col_space_multiplier: Number of spaces between columns relative to the mean width of
    the characters on the page. Default is 3.0.
    :param col_marker_line: Line that dictates the positions of the columns. Default is None - which means the columns
    will be identified based on the spatial values of the nodes.
    :param insert_col_before: If set to True, the code will insert an empty column header (col 0)
    based on col_marker_line. Default is False.
    :param insert_col_after: If set to True, the code will append an empty column header (last column)
    based on col_marker_line. Default is False.
    :param insert_col_index: Index where an empty column will be inserted in col_marker_line. Default is None.
    :param use_graphical_nodes: If set to True, th e graphical figures (lines and rects) that are identified on
    the page will be used to identify the columns of the table. Default is False.
    :param graphic_slop: The value of the slop allowed in lining up the nodes. Default is 1.0 which means it is
    +/- 1.0 of the mean character width of the line. This is only checked when use_graphical_nodes is set to True.
    """

    selector_str = "//*[hasTag('" + tag_name + "')]"
    transform_lines_to_table(node, tag_name, node.select(selector_str), col_space_multiplier=col_space_multiplier,
                             col_marker_line=col_marker_line, insert_col_before=insert_col_before,
                             insert_col_after=insert_col_after, insert_col_index=insert_col_index,
                             use_graphical_nodes=use_graphical_nodes, graphic_slop=graphic_slop)


def transform_lines_to_table(node, tag_name, table_lines,
                             col_space_multiplier=3.0, col_marker_line=None,
                             insert_col_before=False, insert_col_after=False,
                             insert_col_index=None,
                             use_graphical_nodes=False, graphic_slop=10):
    """
    Transforms each line into columns first and then groups them according to their positions.
    Note that this transforms the document.

    :param tag_name: The tag name used for the table rows
    :param table_lines: The lines that will be transformed into table
    :param col_space_multiplier: Number of spaces between columns relative to the mean width of
    the characters on the page. Default is 3.0.
    :param col_marker_line: Line that dictates the positions of the columns. Default is None - which means the columns
    will be identified based on the spatial values of the nodes.
    :param insert_col_before: If set to True, the code will insert an empty column header (col 0)
    based on col_marker_line. Default is False.
    :param insert_col_after: If set to True, the code will append an empty column header (last column)
    based on col_marker_line. Default is False.
    :param insert_col_index: Index where an empty column will be inserted in col_marker_line. Default is None.
    :param use_graphical_nodes: If set to True, th e graphical figures (lines and rects) that are identified on
    the page will be used to identify the columns of the table. Default is False.
    :param graphic_slop: The value of the slop allowed in lining up the nodes. Default is 1.0 which means it is
    +/- 1.0 of the mean character width of the line. This is only checked when use_graphical_nodes is set to True.
    """
    if len(table_lines) == 0:
        return

    # If col_marker_line is given, identify the column positions from this line
    if col_marker_line:
        transform_line_to_columns(col_marker_line, col_space_multiplier,
                                  use_graphical_nodes=use_graphical_nodes,
                                  graphic_slop=graphic_slop)

        if insert_col_before or insert_col_after or insert_col_index is not None:
            # Insert an empty column in col_marker_line
            insert_col_before_or_after(col_marker_line, insert_col_before, insert_col_after,
                                       insert_col_index, col_space_multiplier, use_graphical_nodes)

    # Check first if the line is not transformed to columns yet
    if table_lines[0].get_children()[0].node_type == 'word':
        transform_line_to_columns(table_lines[0], col_space_multiplier, col_marker_line,
                                  use_graphical_nodes=use_graphical_nodes,
                                  graphic_slop=graphic_slop)

    table = [] if col_marker_line else [table_lines[0].get_children().copy()]

    line_start_index = 0 if col_marker_line else 1
    for line in table_lines[line_start_index:]:
        # Check first if the line is not transformed to columns yet
        if line.get_children()[0].node_type == 'word':
            transform_line_to_columns(line, col_space_multiplier, col_marker_line,
                                      use_graphical_nodes=use_graphical_nodes,
                                      graphic_slop=graphic_slop)

        # Align the columns of this line with the rest from the table
        if col_marker_line:
            temp_line_columns = adjust_col_marker_line_columns(line, col_marker_line)
        else:
            temp_line_columns = adjust_table_line_columns(line, table, col_space_multiplier)
        table.append(temp_line_columns)

    # Update the bbox of each column nodes based on min_x1 and max_x2 for each column
    # If there is/are inserted columns in col_marker_line, we should only get the x1 and x2 of those columns with text
    min_x1_list = []
    max_x2_list = []
    for col_idx in range(len(table[0])):
        x1_list = [row[col_idx].get_bbox()[0] for row in table if len(row[col_idx].get_all_content()) > 0]
        x2_list = [row[col_idx].get_bbox()[2] for row in table if len(row[col_idx].get_all_content()) > 0]

        min_x1_list.append(min(x1_list) if x1_list else
                           col_marker_line.get_children()[col_idx].get_bbox()[0] if
                           col_marker_line and len(col_marker_line.get_children()) > col_idx else None)
        max_x2_list.append(max(x2_list) if x2_list else
                           col_marker_line.get_children()[col_idx].get_bbox()[2] if
                           col_marker_line and len(col_marker_line.get_children()) > col_idx else None)

        # Remove None if any
        min_x1_list = [x1 for x1 in min_x1_list if x1 is not None]
        max_x2_list = [x2 for x2 in max_x2_list if x2 is not None]

    # Need to add these table columns as children of each column line
    for row_index, table_row in enumerate(table):
        for col_idx, col in enumerate(table_row):
            # Tag each column
            col.tag(f"{tag_name}/col{col_idx}")
            col.set_bbox([min_x1_list[col_idx] if len(min_x1_list) > col_idx else col.get_bbox()[0],
                          col.get_bbox()[1],
                          max_x2_list[col_idx] if len(max_x2_list) > col_idx else col.get_bbox()[2],
                          col.get_bbox()[3]])

        table_line = table_lines[row_index]
        table_line.adopt_children(table_row, replace=True)


def insert_col_before_or_after(node, insert_col_before, insert_col_after, insert_col_index,
                               col_space_multiplier, use_graphical_nodes):
    # Insert (an) empty column/s in col_marker_line
    # insert_col_index is expected to be the index after column before/after is/are added

    new_columns = node.get_children()
    if insert_col_before:
        # First check if the empty column has already been inserted
        # Happens during the automatic table tagger (since the algo is run twice on the same document for testing)
        if not node.get_children()[0].get_all_content():
            return

        if use_graphical_nodes:
            # Adjust the column's x first
            graphical_nodes = [gn for gn in node.select('parent::page')[0].select('//rect | //figure-line')
                               if gn.get_bbox()[0] < node.get_x()]

            # Get the line with the max x from graphical_nodes
            max_x = max([gn.get_bbox()[0] for gn in graphical_nodes])
            first_col_node = node.get_children()[0]
            first_col_node.set_bbox([max_x, node.get_y(), first_col_node.get_bbox()[2], node.get_bbox()[3]])
            node.set_bbox([max_x, node.get_y(), node.get_bbox()[2], node.get_bbox()[3]])

        new_x2 = node.get_x() - 0.01 if use_graphical_nodes \
            else node.get_x() - col_space_multiplier * node.get_statistics()['updated_mean_width']
        column_node = node.document.create_node(node_type='column')
        column_node.set_bbox([0.0, node.get_y(), new_x2, node.get_bbox()[3]])
        new_columns.insert(0, column_node)

    if insert_col_after:
        if not node.get_children()[-1].get_all_content():
            return

        if use_graphical_nodes:
            # Adjust the column's x first
            graphical_nodes = [gn for gn in node.select('parent::page')[0].select('//rect | //figure-line')
                               if gn.get_bbox()[0] > node.get_bbox()[2]]

            # Get the line with the max x from graphical_nodes
            min_x = min([gn.get_bbox()[0] for gn in graphical_nodes])
            last_col_node = node.get_children()[-1]
            last_col_node.set_bbox([last_col_node.get_x(), node.get_y(), min_x, node.get_bbox()[3]])
            node.set_bbox([node.get_x(), node.get_y(), last_col_node.get_bbox()[2], node.get_bbox()[3]])

        # New bbox is from last node's x + page's width
        new_x1 = node.get_bbox()[2] + 0.01 if use_graphical_nodes \
            else node.get_bbox()[2] + col_space_multiplier * node.get_statistics()['updated_mean_width']
        column_node = node.document.create_node(node_type='column')
        column_node.set_bbox([new_x1, node.get_y(),
                              node.select_first('parent::page').get_bbox()[2], node.get_bbox()[3]])
        new_columns.append(column_node)

    if insert_col_index is not None and 0 < insert_col_index < len(new_columns) - 1:
        # First check if the empty column has already been inserted
        # Happens during the automatic table tagger (since the algo is run twice on the same document for testing)
        if not node.get_children()[insert_col_index].get_all_content():
            return

        # New bbox is +/- col_cpace_multiplier from the columns before and after
        bbox_before = node.get_children()[insert_col_index - 1].get_bbox()
        bbox_after = node.get_children()[insert_col_index].get_bbox()
        x1 = bbox_before[2]
        x2 = bbox_after[0]
        new_x1 = x1 + col_space_multiplier * node.get_statistics()['updated_mean_width']
        new_x2 = x2 - col_space_multiplier * node.get_statistics()['updated_mean_width']
        column_node = node.document.create_node(node_type='column')
        column_node.set_bbox([new_x1, node.get_y(),
                              new_x2, node.get_bbox()[3]])

        new_columns.insert(insert_col_index, column_node)

    node.adopt_children(new_columns, replace=True)
    node.set_bbox_from_children()


def adjust_col_marker_line_columns(node, col_marker_line):
    # Check the other rows of the table if they align with the ref columns
    line_columns = node.get_children().copy()
    ref_columns = col_marker_line.get_children().copy()
    temp_line_columns = []
    ref_col_idx = 0

    if len(line_columns) == len(ref_columns):
        for line_index, line_col in enumerate(line_columns):
            ref_col = ref_columns[line_index]
            temp_line_columns.append(line_col)
            update_bbox_for_columns(ref_col, temp_line_columns[-1], update_col1=False)
        return temp_line_columns

    for line_col_idx, line_col in enumerate(line_columns):
        line_col_bbox = line_col.get_bbox()
        while ref_col_idx < len(ref_columns):
            ref_col = ref_columns[ref_col_idx]

            if (overlaps_with(line_col, ref_col) or line_col.get_x() < ref_col.get_x()) and \
                    ((ref_col_idx + 1 == len(ref_columns) or
                      (ref_col_idx + 1 < len(ref_columns) and (
                              not overlaps_with(line_col, ref_columns[ref_col_idx + 1]) or
                              width_of_overlap(line_col, ref_col) > width_of_overlap(line_col,
                                                                                     ref_columns[ref_col_idx + 1]))))):
                # If line_col overlaps with ref_col and line_col does not overlap with the next one
                temp_line_columns.append(line_col)
                update_bbox_for_columns(ref_col, temp_line_columns[-1], update_col1=False)
                ref_col_idx += 1
                break
            elif ref_col_idx + 1 < len(ref_columns) and \
                    not (overlaps_with(line_col, ref_columns[ref_col_idx + 1])) and \
                    line_col.get_x() + line_col.get_width() < ref_columns[ref_col_idx + 1].get_x():
                # If line_col does not overlap with the next ref_col but its x is less than the next ref_col's x,
                # then line_col lines up with ref_col
                temp_line_columns.append(line_col)
                update_bbox_for_columns(ref_col, temp_line_columns[-1], update_col1=False)
                ref_col_idx += 1
                break
            elif ref_col_idx == len(ref_columns) - 1 and \
                    line_col.get_x() > ref_columns[ref_col_idx].get_x() + ref_columns[ref_col_idx].get_width():
                # If line_col is after the last ref_col, then line_col lines up with ref_col
                temp_line_columns.append(line_col)
                update_bbox_for_columns(ref_col, temp_line_columns[-1], update_col1=False)
                ref_col_idx += 1
                break
            else:
                # Else, Insert an empty column node in temp_line_columns,
                # then check the next ref_col
                column_node = node.document.create_node(node_type='column')
                # Need to get the y values from this row
                ref_col_bbox = ref_col.get_bbox()
                column_node.set_bbox([ref_col_bbox[0], line_col_bbox[1],
                                      ref_col_bbox[2], line_col_bbox[3]])
                temp_line_columns.insert(ref_col_idx, column_node)
                ref_col_idx += 1

    while ref_col_idx < len(ref_columns):
        # Need to add columns since there are more columns in col_marker_line than this line
        ref_col = ref_columns[ref_col_idx]
        ref_col_bbox = ref_col.get_bbox()
        column_node = node.document.create_node(node_type='column')
        column_node.set_bbox([ref_col_bbox[0], line_col_bbox[1],
                              ref_col_bbox[2], line_col_bbox[3]])
        temp_line_columns.insert(ref_col_idx, column_node)
        ref_col_idx += 1

    return temp_line_columns


def adjust_table_line_columns(node, table, col_space_multiplier):
    # Check the other rows of the table if they align with the ref columns
    line_columns = node.get_children().copy()
    ref_columns = table[0]
    temp_line_columns = []  # If there are ones to combine
    ref_col_idx = 0
    overlap_found = False
    mean_width = node.get_statistics()['updated_mean_width']

    for line_col_idx, line_col in enumerate(line_columns):
        line_col_bbox = line_col.get_bbox()
        while ref_col_idx < len(ref_columns):
            ref_col = ref_columns[ref_col_idx]

            if overlaps_with(line_col, ref_col) or \
                    abs(ref_col.get_x() - line_col.get_x()) <= col_space_multiplier * mean_width:
                # If line_col overlaps with ref_col or it is within the allowed col_space_multiplier
                # Check if there are already columns that overlap with this index
                if len(temp_line_columns) > ref_col_idx:
                    # Extend the x value to cover both nodes
                    column_node = temp_line_columns[ref_col_idx]
                    for child in line_col.get_children().copy():
                        column_node.add_child(child)

                    column_node_bbox = column_node.get_bbox()
                    column_node.set_bbox([column_node_bbox[0], column_node_bbox[1],
                                          line_col_bbox[2], line_col_bbox[3]])
                    temp_line_columns[ref_col_idx] = column_node
                    if ref_col_idx < len(ref_columns) - 1 and \
                            overlaps_with(column_node, ref_columns[ref_col_idx + 1]):
                        update_bbox_for_columns(ref_col, temp_line_columns[ref_col_idx], update_col1=False)
                    else:
                        update_bbox_for_columns(ref_col, temp_line_columns[ref_col_idx], update_col1=True)

                else:
                    temp_line_columns.append(line_col)
                    update_bbox_for_columns(ref_col, temp_line_columns[-1], update_col1=True)

                overlap_found = True
                break

            # Do not overlap
            elif line_col.get_x() < ref_col.get_x():
                # Check if the line_col's x is before ref_col's x
                # Insert an 'empty' column node in all the rows in the table,
                # then check the next line_col
                for row in table:
                    # Need to get the y values from this row
                    column_node = node.document.create_node(node_type='column')
                    row0_bbox = row[0].get_bbox()
                    column_node.set_bbox([line_col_bbox[0], row0_bbox[1],
                                          line_col_bbox[2], row0_bbox[3]])
                    row.insert(ref_col_idx, column_node)

                # Add the line_col to temp_line_columns
                temp_line_columns.append(line_col)

                # Updating ref_col_idx since we inserted a column
                ref_col_idx += 1
                overlap_found = False
                break

            # Check if the line_col's x is after ref_col's x + width
            elif line_col.get_x() > ref_col.get_x() + ref_col.get_width() + col_space_multiplier * mean_width:
                if ref_col_idx == len(ref_columns) - 1:
                    if overlap_found:
                        # These are extra columns
                        temp_line_columns.append(line_col)
                        # Append an empty column node to all the rows in the table
                        for row in table:
                            column_node = node.document.create_node(node_type='column')
                            # Need to get the y values from this row
                            row0_bbox = row[0].get_bbox()
                            column_node.set_bbox([line_col_bbox[0], row0_bbox[1],
                                                  line_col_bbox[2], row0_bbox[3]])
                            row.append(column_node)

                        ref_col_idx += 1
                        overlap_found = True
                        break

                    elif line_col_idx == len(line_columns) - 1:
                        # This is the last index, put line_col in the last index
                        temp_line_columns.append(line_col)

                        overlap_found = True
                        break

                    else:
                        # We will not break since we want to keep this line_col
                        # This is not the last col for line columns so we will just say overlap is found
                        # since we have already seen all the ref_cols
                        # Insert an empty column in temp_line_columns
                        column_node = node.document.create_node(node_type='column')
                        # Need to get the y values from this row
                        ref_col_bbox = ref_col.get_bbox()
                        column_node.set_bbox([ref_col_bbox[0], line_col_bbox[1],
                                              ref_col_bbox[2], line_col_bbox[3]])
                        temp_line_columns.insert(ref_col_idx, column_node)

                        overlap_found = True

                elif overlap_found:
                    # If overlap is found for ref_col, move on to the next ref_col
                    ref_col_idx += 1
                    overlap_found = False

                else:
                    # Else, Insert an empty column node in temp_line_columns,
                    # then check the next ref_col
                    column_node = node.document.create_node(node_type='column')
                    # Need to get the y values from this row
                    ref_col_bbox = ref_col.get_bbox()
                    column_node.set_bbox([ref_col_bbox[0], line_col_bbox[1],
                                          ref_col_bbox[2], line_col_bbox[3]])
                    ref_col_idx += 1
                    temp_line_columns.insert(ref_col_idx, column_node)
                    overlap_found = False

            # Append this column_node (with the right text) to temp_line_columns
            else:
                temp_line_columns.append(line_col)
                ref_col_idx += 1
                overlap_found = False

    for idx in range(len(ref_columns) - len(temp_line_columns)):
        # If there are less columns in the given line compared to the reference
        column_node = node.document.create_node(node_type='column')
        # Put a high x1 so it does not get read as a min
        # Put a low x2 so it does not get read as a max
        column_node.set_bbox([10000.00, line_col_bbox[1],
                              0.00, line_col_bbox[3]])
        temp_line_columns.append(column_node)

    return temp_line_columns


def update_bbox_for_columns(col1, col2, update_col1=True):
    """
    Updates the bounding box of the columns when lining them up in a table
    :param col1:  column 1
    :param col2: column 2
    :param update_col1: if set to False, only col2's bbox will be updated. Default is True.
    """
    # Set the bounding box to cover the min x1 and max x2
    min_x1 = min(col1.get_bbox()[0], col2.get_bbox()[0])
    max_x2 = max(col1.get_bbox()[2], col2.get_bbox()[2])

    if update_col1:
        col1_new_bbox = [min_x1, col1.get_bbox()[1], max_x2, col1.get_bbox()[3]]
        col1.set_bbox(col1_new_bbox)

    col2_new_bbox = [min_x1, col2.get_bbox()[1], max_x2, col2.get_bbox()[3]]
    col2.set_bbox(col2_new_bbox)


class DataMarker:
    data_marker_text: Optional[str] = ''
    data_marker_text_re: Optional[str] = ''
    data_marker_bbox: Optional[List[float]]
    data_value_bbox: Optional[List[float]]
    data_value_direction: Optional[str] = ''
    data_type: Optional[str] = ''


def get_data_marker_column_and_index(data_marker_line: ContentNode, data_marker: DataMarker):
    # Get column and column index where the data marker is and the column overlaps with the marker (x-axis)
    try:
        column_index = [col_idx for col_idx, column in enumerate(data_marker_line.select('//column'))
                        if re.match(f'(?i){data_marker.data_marker_text_re}', column.get_all_content()) and
                        percent_nodes_overlap(column.get_bbox(), data_marker.data_marker_bbox,
                                              axis_overlap='x') >= MIN_OVERLAP_PERCENTAGE_X][0]

    except IndexError:
        return None, None

    return data_marker_line.select('//column')[column_index], column_index


def data_marker_overlaps_with_target_marker(data_marker_line_bbox, data_word_bbox, target_data_marker: DataMarker,
                                            overlap_percentage=OVERLAP_PERCENTAGE):
    # Checks that data_marker_line overlaps with the bbox provided in the template
    template_data_marker_bbox = target_data_marker.data_marker_bbox
    template_data_value_bbox = target_data_marker.data_value_bbox
    return (percent_nodes_overlap(data_marker_line_bbox, template_data_marker_bbox, 'x') >= overlap_percentage
            and percent_nodes_overlap(data_word_bbox, template_data_value_bbox, 'x') >= overlap_percentage)


def get_column_below_or_above_data(data_marker_line: ContentNode, data_marker: DataMarker,
                                   target_lines: List[ContentNode],
                                   col_space_multiplier=2.0, column_direction='column_below'):
    data_marker_line_index = target_lines.index(data_marker_line)
    logger.info("get_column_below_or_above_data: %s", data_marker)

    # Can't get a next line
    if len(target_lines) <= data_marker_line_index:
        logger.info(f'No next line for data marker {data_marker.data_marker_text_re} in line {data_marker_line_index}')
        return None

    # Check if the next line overlaps with the y value of data_value_bbox)
    # Make the words as children again of this line
    for count in range(1, 4):
        if len(target_lines) <= data_marker_line_index + count:
            return None
        try:
            next_line = target_lines[data_marker_line_index + count] if column_direction == 'column_below' \
                else target_lines[data_marker_line_index - count]
        except IndexError:
            return None

        # Do not set col_marker_line since the line where the marker is does not guarantee the correct column markers
        [line.adopt_children(line.select('//word'), replace=True) for line in [data_marker_line, next_line]]
        [transform_line_to_columns(tl, col_space_multiplier=col_space_multiplier) for tl in
         [data_marker_line, next_line]]

        column, column_index = get_data_marker_column_and_index(data_marker_line, data_marker)
        if column_index is None:
            return None

        try:
            column_below = [col for col in next_line.select('//column')
                            if percent_nodes_overlap(column.get_bbox(), col.get_bbox(),
                                                     axis_overlap='x') >= 0.4][0]
        except IndexError:
            logger.info("No column below found")
            column_below = None

        if column and column_below and column_below.get_all_content() and \
                data_marker_overlaps_with_target_marker(column.get_bbox(), column_below.get_bbox(), data_marker):
            # This is the data found in the data_marker_line
            return column_below

    return None



import numpy as np
from kodexa import Document
from kodexa.spatial.bbox_common import percent_nodes_overlap

KDXA_BBOX_KEY = 'kodexa_bbox'


def create_page_node_keep_azure_lines(document, page_node, azure_page, overlap_percentage=0.6):
    page_node = document.create_node("content-area", parent=page_node)
    page_node.set_bbox([0, 0, azure_page['width'], azure_page['height']])

    # Now check all the lines identified by Form Recognizer
    all_words = get_azure_page_words(azure_page)

    included_words = []
    word_index = 0
    for line_index, line in enumerate(azure_page['lines']):
        # Create the line and set bbox
        line_node = document.create_node("line", parent=page_node)
        line_node.set_bbox(convert_azure_bbox(line, azure_page))
        line_words = []
        while True and word_index < len(all_words):
            if azure_kdxa_nodes_overlap(all_words[word_index], line_node, azure_page, overlap_percentage):
                line_words.append(all_words[word_index])
                included_words.append(all_words[word_index])
                word_index += 1
            else:
                break

        for word in line_words:
            word_node = document.create_node("word", parent=line_node)
            word_node.set_bbox(convert_azure_bbox(word, azure_page))
            word_node.set_content_parts([word['content']])

    # No issue
    return False


def create_kddb_from_azure(azure_data, keep_azure_lines=True, overlap_percentage=0.6):
    """

    :param azure_data: dictionary of the json data returned by Azure
    :param keep_azure_lines: If set to False, the lines returned will match how Kodexa put lines together.
    Default is True - keeping how Azure creates lines.
    :param overlap_percentage: the percentage of overlap for nodes to be considered on the same line (from 0.0 to 1.0)
    Default is 0.6.
    :return: kddb document with the content nodes
    """
    # This restores the lines provided by Azure (not matching Kodexa)
    if not azure_data:
        return None

    # Create an empty document
    document = Document()
    root_node = document.create_node("root")

    for page_index, azure_page in enumerate(azure_data['analyzeResult']['pages']):
        page_node = document.create_node("page", parent=root_node)
        page_node.set_bbox([0, 0, azure_page['width'], azure_page['height']])
        if 'angle' in azure_page.keys() and azure_page['angle'] is not None:
            page_node.add_feature('page', 'angle', azure_page['angle'])

        # Create the kddb document (keeping Azure lines or matching them with Kodexa lines)
        issue_found = create_page_node_keep_azure_lines(document, page_node, azure_page,
                                                        overlap_percentage=overlap_percentage) \
            if keep_azure_lines else create_page_node_line_up_kodexa(document, page_node, azure_page)

        if issue_found:
            return None



    document.content_node = root_node
    document.add_mixin('spatial')

    return document


def create_page_node_line_up_kodexa(document, page_node, azure_page):
    # Group Azure's OCR words into lines to match how Kodexa sees lines
    content_area_node = document.create_node("content-area", parent=page_node)
    content_area_node.set_bbox([0, 0, azure_page['width'], azure_page['height']])

    line_groups = group_azure_lines_to_kodexa_lines(azure_page)
    azure_page_words = get_azure_page_words(azure_page)
    already_added_azure_words = []

    for line_group_index, line_group in enumerate(line_groups):
        # Sort words in word_group based on x
        line_group = sorted(line_group, key=lambda d: d[KDXA_BBOX_KEY][0])

        # Create the line and set bbox
        line_node = document.create_node("line", parent=content_area_node)

        char_count = 0
        chars_total_width = 0.0

        for azure_line in line_group:
            # Now create a line for each line_group and get the words
            try:
                azure_line_words = azure_line['words']
            except KeyError:
                # The azure line does not contain the words in the metadata - so we have to find them
                azure_line_words = get_azure_line_words(azure_line, azure_page, azure_page_words,
                                                        already_added_azure_words)

            if not azure_line_words:
                continue

            for word in azure_line_words:
                word_node = document.create_node("word", parent=line_node)
                word_node.set_bbox(convert_azure_bbox(word, azure_page))

                if 'confidence' in word.keys():
                    word_node.add_feature('azure', 'confidence', word['confidence'])

                word_node.set_content_parts([word['content']])
                chars_total_width += word_node.get_width()
                char_count += len(word['content'])

        line_node.set_bbox_from_children()
        if char_count > 0:
            line_node.set_statistics({'updated_mean_width': chars_total_width / char_count})
        else:
            line_node.set_statistics({'updated_mean_width': 0.0})

    # This is a good kddb document (no issue)
    return False


def get_azure_line_words(azure_line, azure_page, azure_page_words, already_added_azure_words):
    azure_line_words = []
    word_texts = azure_line['content'].split()
    for wt in word_texts:
        try:
            word = [w for w in azure_page_words if w['content'] == wt and
                    azure_nodes_overlap(w, azure_line, azure_page) and w not in already_added_azure_words][0]
        except IndexError:
            print('Issue with converting Azure lines to Kodexa lines')
            continue

        # Add this word as it has already been found
        already_added_azure_words.append(word)

        # line words to return
        azure_line_words.append(word)

    return azure_line_words


def get_azure_page_words(azure_page):
    if 'words' in azure_page.keys():
        page_words = azure_page['words']
    else:
        page_lines = azure_page['lines']
        page_words = []
        [page_words.extend(pl['words']) for pl in page_lines]
        for pw in page_words:
            pw['content'] = pw['text']

    return page_words


def get_azure_next_line(document_lines, ref_line, direction='right', overlap_percentage=0.6):
    if ref_line not in document_lines:
        return None

    ref_bbox = ref_line.get_bbox()
    possible_lines = document_lines[max(document_lines.index(ref_line) - 25, 0):
                                    min(document_lines.index(ref_line) + 25, len(document_lines))]

    if direction == 'right':
        # Get all the lines to the right of the cell, where the y overlaps; and sort according to x
        right_lines = [right_line for right_line in possible_lines
                       if right_line.get_bbox()[0] >= ref_bbox[2] and
                       percent_nodes_overlap(ref_bbox, right_line.get_bbox(), 'y') >= overlap_percentage]

        if not right_lines:
            return None

        # Sort by x (increasing)
        sorted_next_right_lines = [right_lines[0]]
        [sorted_next_right_lines.insert(0, right_line) for right_line in right_lines
         if right_line.get_bbox()[0] < sorted_next_right_lines[0].get_bbox()[0]]

        return sorted_next_right_lines[0]

    elif direction == 'left':
        # Get all the lines to the left of the cell, where the y overlaps; and sort according to x2 (decreasing)
        left_lines = [left_line for left_line in possible_lines
                      if left_line.get_bbox()[2] <= ref_bbox[0] and
                      percent_nodes_overlap(ref_bbox, left_line.get_bbox(), 'y') >= overlap_percentage]

        if not left_lines:
            return None

        # Sort by x (decreasing)
        sorted_next_left_lines = [left_lines[0]]
        [sorted_next_left_lines.insert(0, left_line) for left_line in left_lines
         if left_line.get_bbox()[0] > sorted_next_left_lines[0].get_bbox()[0]]

        return sorted_next_left_lines[0]

    elif direction == 'down':
        # Get all the lines below of the cell, where the x overlaps; and sort according to y
        down_lines = [down_line for down_line in possible_lines if
                      down_line.get_bbox()[3] <= ref_bbox[1] and
                      percent_nodes_overlap(ref_bbox, down_line.get_bbox(), 'x') >= overlap_percentage]

        if not down_lines:
            return None

        # Sort by y (increasing since 0 is at the bottom of the page)
        sorted_next_down_lines = [down_lines[0]]
        [sorted_next_down_lines.insert(0, down_line) for down_line in down_lines
         if down_line.get_bbox()[1] > sorted_next_down_lines[0].get_bbox()[1]]

        return sorted_next_down_lines[0]

    elif direction == 'up':
        # Get all the lines above of the cell, where the x overlaps; and sort according to y
        up_lines = [up_line for up_line in possible_lines if
                    up_line.get_bbox()[1] >= ref_bbox[3] and
                    percent_nodes_overlap(ref_bbox, up_line.get_bbox(), 'x') >= overlap_percentage]

        if not up_lines:
            return None

        # Sort by y (decreasing since 0 is at the bottom of the page)
        sorted_next_up_lines = [up_lines[0]]
        [sorted_next_up_lines.insert(0, up_line) for up_line in up_lines
         if up_line.get_bbox()[1] < sorted_next_up_lines[0].get_bbox()[1]]

        return sorted_next_up_lines[0]

    elif direction == 'up_left':
        # Get all the lines above of the cell, where the x is to the left of the cell
        up_left_lines = [up_line for up_line in possible_lines if
                         up_line.get_bbox()[1] >= ref_bbox[3] and
                         ref_bbox[2] > up_line.get_x() and ref_bbox[0] - up_line.get_bbox()[2] <= 0.75]

        if not up_left_lines:
            return None

        # Sort by y (decreasing since 0 is at the bottom of the page)
        sorted_next_up_left_lines = [up_left_lines[0]]
        [sorted_next_up_left_lines.insert(0, up_left_line) for up_left_line in up_left_lines
         if up_left_line.get_bbox()[1] < sorted_next_up_left_lines[0].get_bbox()[1]]

        return sorted_next_up_left_lines[0]

    return None


def group_azure_lines_to_kodexa_lines(page):
    # Group the azure lines into Kodexa lines
    # Each line group is a list of lines
    page_lines = page['lines']
    if not page_lines:
        return []

    # Convert Azure lines into Kodexa lines
    line_heights = []
    char_widths = []

    for page_line in page_lines:
        page_line[KDXA_BBOX_KEY] = convert_azure_bbox(page_line, page)
        page_line['content'] = page_line['content'] if 'content' in page_line.keys() \
            else page_line['text'] if 'text' in page_line.keys() else None

    # heights = np.array([page_line[KDXA_BBOX_KEY][3] - page_line[KDXA_BBOX_KEY][1] for page_line in page_lines])
    # stdev_height = np.std(heights)
    #
    # # Get the lines with height above the standard deviation
    # lines_to_separate = [page_lines[i] for i in range(len(line_heights)) if line_heights[i] >= stdev_height]
    lines_to_separate = []
    # # Initialize line_groups
    # line_groups = [[line] for line in lines_to_separate]
    line_groups = []

    # Sort the remaining page_lines by y1
    lines_to_sort = sorted([page_line for page_line in page_lines if page_line not in lines_to_separate],
                           key=lambda d: d[KDXA_BBOX_KEY][1], reverse=True)

    line_groups.extend(check_azure_line_group(lines_to_sort))

    # Need to sort the line_groups based on y
    line_groups.sort(key=lambda line_group: [line[KDXA_BBOX_KEY][1] for line in line_group], reverse=True)
    return line_groups


def get_bbox_of_line_group(line_group):
    if not line_group:
        return None

    # Gets the bounding box of the line_group
    x1_values = [line[KDXA_BBOX_KEY][0] for line in line_group]
    y1_values = [line[KDXA_BBOX_KEY][1] for line in line_group]
    x2_values = [line[KDXA_BBOX_KEY][2] for line in line_group]
    y2_values = [line[KDXA_BBOX_KEY][3] for line in line_group]

    return [min(x1_values), min(y1_values), max(x2_values), max(y2_values)]


def page_line_overlaps_with_line_group(page_line_dict, line_group, next_line_group,
                                       axis_overlap='y', min_overlap_percentage_y=0.4):
    page_line_dict_bbox = page_line_dict[KDXA_BBOX_KEY]
    # line_group_bbox = get_bbox_of_line_group(line_group)
    # next_line_group_bbox = get_bbox_of_line_group(next_line_group)
    line_group_bbox = line_group[-1][KDXA_BBOX_KEY]
    next_line_group_bbox = next_line_group[-1][KDXA_BBOX_KEY] if next_line_group else None

    if next_line_group:
        # Check if page_line_dict overlaps with both line_group and next_line_group
        # If it does, only return true if the overlap is more for line_group than next_line_group
        line_group_overlap = percent_nodes_overlap(page_line_dict_bbox, line_group_bbox, axis_overlap=axis_overlap)
        next_line_group_overlap = percent_nodes_overlap(page_line_dict_bbox, next_line_group_bbox,
                                                        axis_overlap=axis_overlap)
        if line_group_overlap >= min_overlap_percentage_y and \
                next_line_group_overlap >= min_overlap_percentage_y and \
                node_heights_are_valid(page_line_dict_bbox, line_group) and \
                node_heights_are_valid(page_line_dict_bbox, next_line_group):
            return page_line_dict[KDXA_BBOX_KEY][0] - line_group_bbox[0] < \
                page_line_dict[KDXA_BBOX_KEY][0] - next_line_group_bbox[0]
        elif line_group_overlap >= min_overlap_percentage_y and \
                node_heights_are_valid(page_line_dict_bbox, line_group):
            return True
        else:
            return False
    else:
        return percent_nodes_overlap(page_line_dict_bbox, line_group_bbox, axis_overlap=axis_overlap) \
            >= min_overlap_percentage_y and \
            node_heights_are_valid(page_line_dict_bbox, line_group)


def node_heights_are_valid(new_bbox, line_group):
    # The font size/height of the nodes should be within 60% of each other
    # Get the min_height and max_height of each line
    # return True
    height_values = [line[KDXA_BBOX_KEY][3] - line[KDXA_BBOX_KEY][1] for line in line_group]
    min_height = min(height_values)
    max_height = max(height_values)
    new_height = new_bbox[3] - new_bbox[1]
    return min(new_height, min_height) / max(new_height, min_height) >= 0.55 or \
        min(new_height, max_height) / max(new_height, max_height) >= 0.55


def check_azure_line_group(original_line_group):
    # Confirms that the azure lines in this group do not overlap on the x-axis.
    # If they do, then we break down this line group into multiple groups
    if not original_line_group:
        return []

    # Sort the lines in original_line_group by x
    original_line_group = sorted(original_line_group, key=lambda d: d[KDXA_BBOX_KEY][0])

    line_groups = [[original_line_group[0]]]
    for azure_line in original_line_group[1:]:
        line_group_inserted = False
        for lg_index, line_group in enumerate(line_groups):
            # If the azure line does not x-overlap with this line group, check if they y-overlap (+/- 0.1)
            if (azure_line[KDXA_BBOX_KEY][0] >= line_group[-1][KDXA_BBOX_KEY][2] - 0.1 or
                azure_line[KDXA_BBOX_KEY][2] <= line_group[0][KDXA_BBOX_KEY][0] + 0.1) and \
                    page_line_overlaps_with_line_group(azure_line, line_group,
                                                       line_groups[lg_index + 1] if lg_index + 1 < len(line_groups)
                                                       else None):
                line_group.append(azure_line)
                line_group_inserted = True
                break

        if line_group_inserted:
            continue

        line_groups.append([azure_line])
        line_groups.sort(key=lambda line_group: [line[KDXA_BBOX_KEY][1] for line in line_group], reverse=True)

    return line_groups


def convert_azure_bbox(azure_obj, azure_page):
    # boundingBox has 8 points
    if 'boundingBox' in azure_obj:
        bbox_points = azure_obj['boundingBox']
    elif 'bounding_box' in azure_obj:
        bbox_points = azure_obj['bounding_box']
    elif 'polygon' in azure_obj:
        bbox_points = azure_obj['polygon']
    else:
        raise Exception('Could not find bounding box in azure object')

    if len(bbox_points) == 8:
        # The x and y are given as a flat data
        x_points = bbox_points[0::2]
        y_points = bbox_points[1::2]
    else:
        # The x and y area given as pairs
        x_points = [pt['x'] for pt in bbox_points]
        y_points = [pt['y'] for pt in bbox_points]

    # Convert these points based on the angle of the page
    if 'angle' in azure_page.keys() and azure_page['angle'] is not None:
        rotated_x_points = []
        rotated_y_points = []
        for (x, y) in [(x_points[index], y_points[index]) for index in range(len(x_points))]:
            rot_x, rot_y = rotate((x, y), (0, 0), azure_page['angle'])
            rotated_x_points.append(rot_x)
            rotated_y_points.append(rot_y)
        x_points = rotated_x_points
        y_points = rotated_y_points

    x1 = min(x_points)
    y1 = azure_page['height'] - max(y_points)
    x2 = max(x_points)
    y2 = azure_page['height'] - min(y_points)

    return [x1, y1, x2, y2]


def percent_azure_nodes_overlap(azure_node1, azure_node2, azure_page, axis_overlap='y'):
    kdxa_bbox1 = convert_azure_bbox(azure_node1, azure_page)
    kdxa_bbox2 = convert_azure_bbox(azure_node2, azure_page)

    return percent_nodes_overlap(kdxa_bbox1, kdxa_bbox2, axis_overlap=axis_overlap)


def azure_kdxa_nodes_overlap(azure_node1, kdxa_node2, azure_page, overlap_percentage=0.6):
    node1_bbox = convert_azure_bbox(azure_node1, azure_page)
    node2_bbox = kdxa_node2.get_bbox()

    return percent_nodes_overlap(node1_bbox, node2_bbox, axis_overlap='x') >= overlap_percentage and \
        percent_nodes_overlap(node1_bbox, node2_bbox, axis_overlap='y') >= overlap_percentage


def azure_nodes_overlap(azure_node1, azure_node2, azure_page, overlap_percentage=0.6):
    return percent_azure_nodes_overlap(azure_node1, azure_node2, azure_page, axis_overlap='x') >= overlap_percentage and \
        percent_azure_nodes_overlap(azure_node1, azure_node2, azure_page, axis_overlap='y') >= overlap_percentage


def rotate(point, origin, degrees):
    radians = np.deg2rad(degrees)
    x, y = point
    offset_x, offset_y = origin
    adjusted_x = (x - offset_x)
    adjusted_y = (y - offset_y)
    cos_rad = np.cos(radians)
    sin_rad = np.sin(radians)
    qx = offset_x + cos_rad * adjusted_x + sin_rad * adjusted_y
    qy = offset_y + -sin_rad * adjusted_x + cos_rad * adjusted_y
    return qx, qy



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kodexa',
 'kodexa.assistant',
 'kodexa.connectors',
 'kodexa.model',
 'kodexa.pipeline',
 'kodexa.platform',
 'kodexa.selectors',
 'kodexa.spatial',
 'kodexa.steps',
 'kodexa.testing',
 'kodexa.training']

package_data = \
{'': ['*']}

install_requires = \
['addict==2.4.0',
 'appdirs>=1.4.4,<2.0.0',
 'better-exceptions>=0.3.3,<0.4.0',
 'datamodel-code-generator>=0.13.0,<0.14.0',
 'deepdiff==5.8.1',
 'jsonpickle==2.2.0',
 'msgpack==1.0.4',
 'ply>=3.11,<4.0',
 'pydantic-yaml>=0.8.0,<0.9.0',
 'pydantic>=1.10.4,<2.0.0',
 'pyfunctional>=1.4.3,<1.5.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.1,<3.0.0',
 'simpleeval==0.9.12',
 'urllib3>=1.26.14,<2.0.0']

setup_kwargs = {
    'name': 'kodexa',
    'version': '6.2.25131219962',
    'description': 'Python SDK for the Kodexa Platform',
    'long_description': '# Kodexa\n\n[![Build and Package with Poetry](https://github.com/kodexa-ai/kodexa/actions/workflows/main.yml/badge.svg?branch=main)](https://github.com/kodexa-ai/kodexa/actions/workflows/main.yml)\n\n![img.png](https://docs.kodexa.com/img.png)\n\nKodexa is a platform for building intelligent document processing pipelines. It is a set of tools and services that\nallow you to build a pipeline that can take a document, extract the content, and then process it to extract the\ninformation you need.\n\nIt is built on a set of core principles:\n\n* **Document Centric** - Kodexa is built around the idea of a document. A document is a collection of content\n  nodes that are connected together. This is a powerful model that allows you to build pipelines that can\n  extract content from a wide range of sources.\n\n* **Pipeline Oriented** - Kodexa is built around the idea of a pipeline. A pipeline is a series of steps that\n  can be executed on a document. This allows you to build a pipeline that can extract content from a wide range\n  of sources.\n\n* **Extensible** - Kodexa is built around the idea of a pipeline. A pipeline is a series of steps that can be executed\n  on a document. This allows you to build a pipeline that can extract content from a wide range of sources.\n\n* **Label Driven** - Kodexa focuses on the idea of labels. Labels are a way to identify content within a document\n  and then use that content to drive the processing of the document.\n\n# Python SDK\n\nThis repository contains the Python SDK for Kodexa. The SDK is the primary way to interact with Kodexa. It allows you to\ndefine actions, models, and pipelines that can be executed on Kodexa. It also includes a complete SDK client for\nworking with a Kodexa platform instance.\n\n## Documentation & Examples\n\nDocumentation is available at the [Kodexa Documentation Portal](https://docs.kodexa.com)\n\n## Current Development\n\n[//]: # (Replace it with the diagrams and descriptions for build releases)\n**BUILD VERSION FLOW**\n![build-version-flow.png](docs%2Fbuild-version-flow.png)\nBuild version will differ based on the branches that are published to pypi.\n\n**GITHUB PROCESS**\n![github-process.png](docs%2Fgithub-process.png)\nChanges that contain bugs, features, and fixes should first be pushed to the test branch. \nOnce these changes are thoroughly tested, they can be submitted as a pull request to the main branch. The pull request should be reviewed and approved by an appropriate person before the changes can be merged.\n\n## Set-up\n\nWe use poetry to manage our dependencies, so you can install them with:\n\n    poetry install\n\nYou can then run the tests with:\n\n    poetry run pytest\n\n# Contributing\n\nWe welcome contributions to the Kodexa platform. Please see our [contributing guide](CONTRIBUTING.md) for more details.\n\n# License\n\nApache 2.0\n\n',
    'author': 'Austin Redenbaugh',
    'author_email': 'austin@kodexa.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
